import orjson
import uuid
import asyncio
from aio_pika import connect, Message
from aio_pika.abc import (
    AbstractChannel,
    AbstractConnection,
    AbstractIncomingMessage,
    AbstractQueue,
)
from typing import MutableMapping
from loguru import logger
from .types import Task, TaskResult


class RPCClient:
    connection: AbstractConnection
    channel: AbstractChannel
    callback_queue: AbstractQueue
    loop: asyncio.AbstractEventLoop

    def __init__(self, amqp_dsn: str, queue_name: str) -> None:
        self.amqp_dsn = amqp_dsn  # RMQ connection string
        self.queue_name = queue_name  # RMQ queue name
        self.futures: MutableMapping[str, asyncio.Future] = {}
        self.loop = asyncio.get_running_loop()

    async def connect(self) -> "RPCClient":
        connection = await connect(self.amqp_dsn)
        self.channel = await connection.channel()
        self.callback_queue = await self.channel.declare_queue(exclusive=True)
        await self.callback_queue.consume(self.on_response)
        return self

    def on_response(self, message: AbstractIncomingMessage) -> None:
        if message.correlation_id is None:
            logger.error(f"Bad message '{message!r}'")
            return
        future: asyncio.Future = self.futures.pop(message.correlation_id)
        future.set_result(message.body)

    async def call(self, task: Task) -> dict:
        correlation_id = str(uuid.uuid4())
        future = self.loop.create_future()
        self.futures[correlation_id] = future
        await self.channel.default_exchange.publish(
            Message(
                orjson.dumps(task.dict()),
                content_type="application/octet-stream",
                correlation_id=correlation_id,
                reply_to=self.callback_queue.name,
            ),
            routing_key=self.queue_name,
        )
        return TaskResult.parse_obj(orjson.loads(await future))



import traceback
from . import logger
from .types import TaskResult, Statuses


def handle_errors(func):
    def wrapper(*args, **akwrgs):
        try:
            return TaskResult(result = func(*args, **akwrgs)).dict()
        except Exception as exc:
            logger.error(f"{exc}")
            return TaskResult(
                status_code = Statuses.err,
                result = {
                    "msg": f"{exc}", 
                    "traceback": traceback.format_exc()
                }
            ).dict()
    return wrapper



import inspect, re
import simplejson
from typing import Tuple, Union
from pydantic import BaseModel
from . import logger
from enum import Enum


class Task(BaseModel):
    procedure: str  # remote procedure name
    kwargs: dict = {}  # remote procedure kwargs

class Statuses(Enum):
    ok:int = 200
    err:int = 500

class TaskResult(BaseModel):
    status_code: Statuses = Statuses.ok
    result:Union[str,dict,None]

    def __str__(self):
        ''' str(TaskResult(result={})) -> '{}' '''
        return simplejson.dumps(self.result)

    class Config:
        use_enum_values = True

class ReturnTypeError(Exception):
    ...


class CallbackHandler:
    """Class to extend"""

    __RETURN_TYPE_TEMPLATE = Tuple[int, dict]
    __TASK_EXECUTOR_PATTERN = "^task_"

    def __init__(self) -> None:
        task_member_count: int = 0  # count of task executor
        for method in inspect.getmembers(self, predicate=inspect.ismethod):
            method_name = method[0]
            if re.match(self.__TASK_EXECUTOR_PATTERN, method_name):
                task_member_count += 1
                if (
                    inspect.signature(getattr(self, method_name)).return_annotation
                    != self.__RETURN_TYPE_TEMPLATE
                ):
                    raise ReturnTypeError(
                        f'"{method_name}" -> Return type must be type <{self.__RETURN_TYPE_TEMPLATE}>'
                    )
        if task_member_count == 0:
            logger.warning("CallbackHandler -> No tasks to execute was founded!")


# ---| Exmaple
# class TestCallbackHandler(CallbackHandler):
#     def __init__(
#         self,
#     ) -> None:
#         super(TestCallbackHandler, self).__init__()

#     def task_test(self, arg1, arg2=0) -> Tuple[int, dict, int]:
#         return 200, {"result": "test({arg1}, {arg2})"}


# TestCallbackHandler()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rpc_call']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=8.3.0,<9.0.0',
 'loguru>=0.6.0,<0.7.0',
 'orjson>=3.8.5,<4.0.0',
 'pika>=1.3.1,<2.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'simplejson>=3.18.1,<4.0.0']

entry_points = \
{'console_scripts': ['client = tests.simple_client:run',
                     'server = tests.simple_server:run']}

setup_kwargs = {
    'name': 'rpc-call',
    'version': '0.1.5',
    'description': '',
    'long_description': '> **Note:**\n> The main idea of the project is based on RabbitMQ RPC, so RabbitMQ must be installed. Please keep it mind.\n\n### Get started\n\n#### *Simple RPC client using FastAPI*\n#### **`client.py`**\n\n```python\nimport asyncio\nfrom fastapi import FastAPI, Depends, status\nfrom rpc_call.async_client import RPCClient\nfrom rpc_call.types import Task, TaskResult\n\napp = FastAPI()\n\nasync def rpc_connection() -> RPCClient:\n    return await RPCClient(\n        amqp_dsn = "amqp://<client>:<passwd>@<host>:<port>/<vhost>",\n        queue_name = "<RabbitMQQueueName>"\n    ).connect()\n\n@app.post("/RPCEndpoint", response_model=TaskResult)\nasync def read_users(task: Task, rpc_conn: RPCClient = Depends(rpc_connection)) -> TaskResult:\n    task_result = await rpc_conn.call(task)\n    return task_result\n```\n```bash\n$ uvicorn client:app --host 0.0.0.0 --port 8000 --reload\n```\n\n#### *Simple RPC server*\n\n#### **`server.py`**\n\n```python\nfrom rpc_call.server import RPCServer\n\nclass CallbackHandler:\n    def test_func(self, arg: str) -> str:\n        return f"test_func(\'{arg}\') call result"\n\nif __name__ == "__main__":\n    RPCServer(\n        amqp_dsn = "amqp://<client>:<passwd>@<host>:<port>/<vhost>",\n        queue_name = "<RabbitMQQueueName>",\n        callback_handler = CallbackHandler\n    )\n```\n```shell\n$ python server.py\n```\n\n### Usage\n\n```bash\ncurl -X \'POST\' \\\n  \'http://127.0.0.1:8000/RPCEndpoint\' \\\n  -H \'accept: application/json\' \\\n  -H \'Content-Type: application/json\' \\\n  -d \'{\n  "procedure": "test_func",\n  "kwargs": {"arg": "Hello world!"}\n}\'\n```\nAnd the response will be:\n```json\n{\n  "status_code": 200,\n  "result": {\n    "test_func": "test_func(\'Hello world!\') call result"\n  }\n}\n```\n\n---\nSo now, with this elegant and simple library, you can call your code remotely! 👏\nGood luck! 😉\n\n\n',
    'author': 'Artyom Syssolov',
    'author_email': 'artyom.syssolov@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pika
import orjson
from loguru import logger
from pydantic import AmqpDsn
from .types import Task, CallbackHandler
from .helpers import handle_errors


class RPCServer:
    def __init__(
        self, callback_handler: CallbackHandler, amqp_dsn: str, queue_name: str
    ) -> None:
        self.callback_handler = callback_handler()  # callback handler class
        self.amqp_dsn = amqp_dsn  # RMQ connection string
        self.queue_name = queue_name  # RMQ queue name
        connection = pika.BlockingConnection(pika.URLParameters(self.amqp_dsn))
        channel = connection.channel()
        # channel.queue_declare(queue='rpc_queue')
        channel.basic_qos(prefetch_count=1)
        channel.basic_consume(
            queue=self.queue_name, on_message_callback=self.on_request
        )
        logger.info(" [x] Awaiting RPC requests")
        channel.start_consuming()

    @handle_errors
    def call_procedure(self, body: bytes):
        task = Task.parse_obj(orjson.loads(body))
        logger.info(f"task -> {task}")
        return getattr(self.callback_handler, task.procedure)(**task.kwargs)

    def on_request(self, ch, method, props, body) -> None:
        response = orjson.dumps(self.call_procedure(body))
        ch.basic_publish(
            exchange="",
            routing_key=props.reply_to,
            properties=pika.BasicProperties(correlation_id=props.correlation_id),
            body=response,
        )
        ch.basic_ack(delivery_tag=method.delivery_tag)


# #
# class TestCallbackHandler:
#     def test(self, arg1, arg2=0):
#         return {'result': 'test({arg1}, {arg2})'}


# def start():
#     RPCServer(callback_handler = TestCallbackHandler())
